# Copyright 2017, 2018, 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import
import subprocess
import yaml
import time
import eventlet
import eventlet.pools
import greenlet
import sys

from netq_lib.orm.redisdb.models import ProcNetDevInterface


global_gthread_dict = {}
global_sub_process = {}
__interface_ethtool_op = {}
__CMD_TIMEOUT = 20


class ReturnGroup:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def stop_command_checker(green_thread, port_name, cmd):
    try:
        (out, err) = green_thread.wait()
        if cmd.returncode != 0:
            __interface_ethtool_op[port_name] = ReturnGroup(success=False, op="")
        else:
            __interface_ethtool_op[port_name] = ReturnGroup(success=True, op=out)
        return
    except greenlet.GreenletExit:
        __interface_ethtool_op[port_name] = ReturnGroup(success=False, op="")
        pass
    except Exception:  # pylint: disable=broad-except
        __interface_ethtool_op[port_name] = ReturnGroup(success=False, op="")
        eventlet.kill(None, *sys.exc_info())


def run_command_safe(command, thread_pool, port_name):
    """Run time-limited command and update the global proc dict __interface_ethtool_op
    with the ethtool status and response
    """
    global global_gthread_dict
    global global_sub_process
    try:
        cmd = subprocess.Popen(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    except (subprocess.CalledProcessError, OSError):
        __interface_ethtool_op[port_name] = ReturnGroup(success=False, op="")
        return

    global_sub_process[port_name] = cmd
    green_thread = thread_pool.spawn(cmd.communicate)
    green_thread.link(stop_command_checker, port_name, cmd)
    global_gthread_dict[port_name] = green_thread


def process_proc_net_dev(output, svc_failed=False, **kwargs):
    """
    Parse /proc/net/dev counters
    """
    global __interface_ethtool_op
    global global_gthread_dict
    global global_sub_process
    __interface_ethtool_op = {}
    global_gthread_dict = {}
    global_sub_process = {}

    command = "ethtool -S "
    thread_pool = kwargs.get('pool', '*')

    with open('/proc/net/dev', 'r') as fh:

        elem_lines = fh.readlines()[2:]

        for line in elem_lines:
            (port_name,
             rx_bytes, rx_packets, rx_errs, rx_drop, rx_fifo, rx_frame, rx_compressed, rx_multicast,
             tx_bytes, tx_packets, tx_errs, tx_drop, tx_fifo, tx_colls, tx_carrier, tx_compressed) = line.strip().split()

            # chop the trailing :
            port_name = port_name[0:-1]

            if port_name.endswith('-v0') or (not (port_name.startswith('swp') or port_name.startswith('eth'))) or ('.' in port_name):
                continue

            run_command_safe(command + port_name, thread_pool, port_name)

        for line in elem_lines:
            (ifname,
             rx_bytes, rx_packets, rx_errs, rx_drop, rx_fifo, rx_frame, rx_compressed, rx_multicast,
             tx_bytes, tx_packets, tx_errs, tx_drop, tx_fifo, tx_colls, tx_carrier, tx_compressed) = line.strip().split()

            # chop the trailing :
            ifname = ifname[0:-1]

            # This is a virtual interface that is created as part of VRR, ignore it
            if ifname.endswith('-v0') or (not (ifname.startswith('swp') or ifname.startswith('eth'))) or ('.' in ifname):
                continue

            endtime = time.time() + int(__CMD_TIMEOUT)

            while True:
                if ifname in __interface_ethtool_op:
                    cmd_out = __interface_ethtool_op.pop(ifname)
                    if cmd_out.success:
                        retcode = 0
                    else:
                        retcode = -1
                    break
                if time.time() > endtime:
                    retcode = -1
                    global_sub_process[ifname].terminate()
                    eventlet.sleep(0.3)
                    global_sub_process[ifname].wait()
                    eventlet.kill(global_gthread_dict[ifname], greenlet.GreenletExit)
                    break
                eventlet.sleep(0.3)

            if retcode == 0:
                ethElems = {}
                try:
                    for ethtoolstat in cmd_out.op.splitlines():
                        if ":" not in line:
                            continue

                        key, value = ethtoolstat.split(':', 1)
                        key = key.strip()
                        value = value.strip()
                        ethElems[key] = value

                except yaml.YAMLError:
                    ethElems = {}

                if('HwIfInOctets' in ethElems and
                        'HwIfInUcastPkts' in ethElems and
                        'HwIfInBcastPkts' in ethElems and
                        'HwIfInMcastPkts' in ethElems and
                        'HwIfOutOctets' in ethElems and
                        'HwIfOutUcastPkts' in ethElems and
                        'HwIfOutMcastPkts' in ethElems and
                        'HwIfOutBcastPkts' in ethElems and
                        'HwIfInDiscards' in ethElems and
                        'HwIfInDot3LengthErrors' in ethElems and
                        'HwIfInErrors' in ethElems and
                        'HwIfOutDiscards' in ethElems and
                        'HwIfOutErrors' in ethElems):

                    iface = create_proc(ifname,
                                        rx_bytes=ethElems['HwIfInOctets'],
                                        rx_drop=ethElems['HwIfInDiscards'],
                                        rx_errs=ethElems['HwIfInDot3LengthErrors'] + ethElems['HwIfInErrors'],
                                        rx_frame=ethElems['HwIfInUcastPkts'] + ethElems['HwIfInBcastPkts'] + ethElems['HwIfInMcastPkts'],
                                        rx_multicast=ethElems['HwIfInMcastPkts'],
                                        rx_packets=ethElems['HwIfInUcastPkts'] + ethElems['HwIfInBcastPkts'] + ethElems['HwIfInMcastPkts'],
                                        tx_bytes=ethElems['HwIfOutOctets'],
                                        tx_drop=ethElems['HwIfOutDiscards'],
                                        tx_errs=ethElems['HwIfOutErrors'],
                                        tx_packets=ethElems['HwIfOutUcastPkts'] + ethElems['HwIfOutMcastPkts'] + ethElems['HwIfOutBcastPkts'],
                                        rx_broadcast=ethElems['HwIfInBcastPkts'],
                                        tx_broadcast=ethElems['HwIfOutBcastPkts'],
                                        tx_multicast=ethElems['HwIfOutMcastPkts'])

                else:
                    iface = create_proc(ifname, rx_bytes=rx_bytes, rx_drop=rx_drop, rx_errs=rx_errs, rx_frame=rx_frame, rx_multicast=rx_multicast, rx_packets=rx_packets,
                                        tx_bytes=tx_bytes, tx_carrier=tx_carrier, tx_colls=tx_colls, tx_drop=tx_drop, tx_errs=tx_errs, tx_packets=tx_packets)
            else:
                iface = create_proc(ifname, rx_bytes=rx_bytes, rx_drop=rx_drop, rx_errs=rx_errs, rx_frame=rx_frame, rx_multicast=rx_multicast, rx_packets=rx_packets,
                                    tx_bytes=tx_bytes, tx_carrier=tx_carrier, tx_colls=tx_colls, tx_drop=tx_drop, tx_errs=tx_errs, tx_packets=tx_packets)

            iface.save()


def create_proc(ifname, rx_bytes=-1, rx_drop=-1, rx_errs=-1, rx_frame=-1, rx_multicast=-1, rx_packets=-1, tx_bytes=-1,
                tx_carrier=-1, tx_colls=-1, tx_drop=-1, tx_errs=-1, tx_packets=-1, rx_broadcast=-1, tx_broadcast=-1, tx_multicast=-1):
    # fifo and compressed related counters are intentionally left out
    iface = ProcNetDevInterface()
    iface.ifname = ifname
    iface.rx_bytes = rx_bytes
    iface.rx_drop = rx_drop
    iface.rx_errs = rx_errs
    iface.rx_frame = rx_frame
    iface.rx_multicast = rx_multicast
    iface.rx_packets = rx_packets
    iface.tx_bytes = tx_bytes
    iface.tx_carrier = tx_carrier
    iface.tx_colls = tx_colls
    iface.tx_drop = tx_drop
    iface.tx_errs = tx_errs
    iface.tx_packets = tx_packets
    iface.rx_broadcast = rx_broadcast
    iface.tx_broadcast = tx_broadcast
    iface.tx_multicast = tx_multicast
    return iface
